/* 
  CESYS Software License - Version 1.0 - January 1st, 2016

  Permission is hereby granted, free of charge, to any person or organization
  obtaining a copy of the software and accompanying documentation covered by
  this license (the "Software") to use, reproduce, display, distribute,
  execute, and transmit the Software, and to prepare derivative works of the
  Software, and to permit third-parties to whom the Software is furnished to
  do so, all subject to the following:

  The software must only be used to operate hardware manufactured by 
  CESYS GmbH, Herzogenaurach.

  The copyright notices in the Software and this entire statement, including
  the above license grant, this restriction and the following disclaimer,
  must be included in all copies of the Software, in whole or in part, and
  all derivative works of the Software, unless such copies or derivative
  works are solely in the form of machine-executable object code generated by
  a source language processor.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
  SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
  FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
  
  Written by Thomas Hoppe <thomas.hoppe@cesys.com>, 2018
*/  

#pragma once

#include <string>
#include <vector>
#include <cstdint>
#include <memory>

// Load C Header into separate namespace.
namespace _c {
	#include "beastlink.h"
}

namespace beastlink {
	class EnumeratedDevice;

	/**
	 * Structure used to transfer version information. Used in
	 * ::getVersion() and DeviceInfo.getFirmwareVersion().
	 */
	struct Version { 
		int major{ 0 }, minor{ 0 }, fix{ 0 }; 

		// Returns version as string.
		std::string toString() const {
			using std::to_string;
			return to_string(major) + "." + to_string(minor) + "." + to_string(fix);
		}
	};

	/**
	 * This type offers access to most of the device information. Some information
	 * is available after opening the device. An instance to this type can either
	 * be retrieved from EnumeratedDevice::getDeviceInfo() or Device::getDeviceInfo().
	 */
	class DeviceInfo {
		friend class Device;
		friend std::vector<EnumeratedDevice> enumerate(uint16_t vid, uint16_t pid);

		_c::BL_DEVICE_HANDLE _handle{ _c::BL_INVALID_DEVICE };
		uint16_t _vid, _pid;

		DeviceInfo(uint16_t vid, uint16_t pid) : _vid(vid), _pid(pid) {}

	public:

		/**
		 * Return vendor ID of device (if applicable).
		 */
		uint16_t getVid() const { return _vid; }

		/**
		 * Return product ID of device (if applicable).
		 */
		uint16_t getPid() const { return _pid; }

		/**
		 * Set user ID. Writes the string to the device EEPROM.
		 * This ID is intended to differentiate multiple devices of the same type connected
		 * to one host. The ID can contain up to 63 ASCII characters.
		 */
		void setUserId(const std::string &userId);

		/**
		 * Return user ID from device EEPROM. See explanation in 
		 * setUserId(). "" if not accessible or in case of error.
		 */
		std::string getUserId() const;

		/**
		 * Return serial number. Not supported by all device 
		 * types / instances. "" if not accessible or in case of error.
		 */
		std::string getSerialNumber() const;

		/**
		 * Return firmware version of device.
		 */
		Version getFirmwareVersion() const;

		/**
		 * Return derivate info string. "" if not accessible or in case of error.
		 * Some devices are available in different versions (different FPGA's, different peripherals).
		 * The returned string contains a textual information about the specific
		 * derivate. For programmatically differentiation, getDerivateId() is better suited.
		 */
		std::string getDerivateInfo() const;

		/**
		 * Return derivate id. 0xffffffff if not accessible or in case of error.
		 * This ID is unique for every derivate of a specific device type. 
		 * Existing derivates can be found in the documentation of the specific device type.
		 */
		uint32_t getDerivateId() const;
	};

	/**
	 * This type offers functionality to communicate with a single device. Instances can be
	 * retrieved from EnumeratedDevice::open() and represent an opened connection to a device.
	 */
	class Device {
		friend class EnumeratedDevice;

		DeviceInfo _info;
		_c::BL_DEVICE_HANDLE _handle;
		bool _closed;

		Device(const DeviceInfo &info, _c::BL_DEVICE_HANDLE handle, bool autoClose) : 
			_info(info), _handle(handle), _closed(!autoClose) {
			_info._handle = handle;
		}

	public:

		/**
		 * Implicitly close device if instance gets destructed.
		 */
		~Device() { try { close(); } catch (std::exception&) {} } // Never throw in dtor.

		/**
		 * This flag can be used in readBlock() and writeBlock() to specify transfer to a 
		 * constant address. The opposite and *default* behavior is incremental address.
		 * Attention: For historical reasons, this flag is interpreted inverse inside the FPGA interface ->
		 * "address auto increment" is 0 if FlagConstAddress is specified.
		 */
		static const int FlagConstAddress = _c::BL_FLAG_CONST_ADDR;

		// *** Generic access.

		/**
		 * Free access to device. After this call the device is available in new enumerations and
		 * this instance of type Device can't be used anymore.
		 */
		void close();

		// *** Data transfer functions.

		/** 
		 * Read 32 bit register value.
		 * @param reg Register in FPGA address space.
		 * @return Value of register.
		 */
		uint32_t readRegister(uint32_t reg);

		/** 
		 * Write 32 bit register value.
		 * @param reg Register in FPGA address space.
		 * @param value Value to write to register.
		 */
		void writeRegister(uint32_t reg, uint32_t value);

		/** 
		 * Read data block.
		 * @param addr Start address in FPGA address space.
		 * @param buffer Space to read to from FPGA address space, must be at least size bytes.
		 * @param size Count of bytes to transfer.
		 * @param flags I/O specific flags (see their individual documentation)
		 *           - Device::FlagConstAddress
		 */
		void readBlock(uint32_t addr, void *buffer, int size, int flags = 0);
		
		/**
		 * Read data block.
		 * @param addr Start address in FPGA address space.
		 * @param buffer Space to read to from FPGA address space.
		 * @param flags I/O specific flags (see their individual documentation)
		 *           - Device::FlagConstAddress
		 */
		void readBlock(uint32_t addr, std::vector<uint8_t> &buffer, int flags = 0);

		/**
		* Read data block.
		* @param addr Start address in FPGA address space.
		* @param bytes Amount of bytes to read.
		* @param flags I/O specific flags (see their individual documentation)
		*           - Device::FlagConstAddress
		*/
		std::vector<uint8_t> readBlock(uint32_t addr, size_t bytes, int flags = 0);

		/** 
		 * Write data block.
		 * @param addr Start address in FPGA address space.
		 * @param buffer Data to write to FPGA address space, must be at least size bytes.
		 * @param size Count of bytes to transfer.
		 * @param flags I/O specific flags (see their individual documentation)
		 *           - Device::FlagConstAddress
		 */
		void writeBlock(uint32_t addr, void *buffer, int size, int flags = 0);

		/**
		 * Write data block.
		 * @param addr Start address in FPGA address space.
		 * @param buffer Data to write to FPGA address space.
		 * @param flags I/O specific flags (see their individual documentation)
		 *           - Device::FlagConstAddress
		 */
		void writeBlock(uint32_t addr, const std::vector<uint8_t> &buffer, int flags = 0);

		// *** FPGA specific.
		
		/** 
		 * Pulse FPGA reset line (the I/O that is chosen to be reset). Useful to bring the FPGA
		 * in a defined state. Implicitly done after FPGA programming.
		 */
		void resetFpga();

		/** 
		 * Configure FPGA with given design in file. Loads a .bin from as specified by 
		 * fileName and programs the FPGA using this design. Subsequently pulses the reset line.
		 * @param fileName File which contains design file data.
		 */
		void programFpgaFromBin(const std::string &fileName);

		/** 
		 * Configure FPGA with given design in memory. Useful when compiling FPGA designs into 
		 * the executable. BoardManager creates sources that can be used this way. 
		 * It is preferred to use programFpgaFromMemoryZ().
		 * Subsequently pulses the reset line.
		 * @param buffer Design pointer.
		 * @param size Len of array in byte.
		 */
		void programFpgaFromMemory(void *buffer, int size, bool compressed);

		// *** Configuration.
	
		/** 
		 * Configure timeout until a pending transfer is stopped and defined as failed.
		 * This timeout is used for all transfers to and from device.
		 * @param timeout Timeout in milliseconds.
		 */ 
		void setTimeout(uint32_t timeout);

		// *** General device info.

		/**
		 * Get general device info.
		 */
		DeviceInfo getDeviceInfo() const { return _info; }
	};

	// When Device is used in different contexts, this is a better way
	// as Device is not copyable.
	using DevicePtr = std::shared_ptr<Device>;

	/**
	 * Entry returned by enumeration, valid only between two
	 * enumerations. Call its open() method to get the
	 * real device handle in form of an instance of type Device.
	 */
	class EnumeratedDevice {
		friend std::vector<EnumeratedDevice> enumerate(uint16_t vid, uint16_t pid);

		DeviceInfo _info;
		int _eid, _erid;

		EnumeratedDevice(const DeviceInfo &info, int eid, int erid) : 
			_info(info), _eid(eid), _erid(erid) {
		}

	public:

		/**
		 * Open this instance, returns object to use for further communication.
		 * An opened device isn't reported in any subsequent enumeration (system wide), until
		 * it's closed.
		 * @param autoClose If true, Device::~Device() will call close().
		 */
		Device open(bool autoClose = true) const;

		/**
		 * Open device as shared ptr, which is better when Device is used
		 * in a non linear context.
		 * @param autoClose If true, Device::~Device() will call close().
		 */
		std::shared_ptr<Device> openAsPtr(bool autoClose = true) const;

		/**
		 * Get general device info.
		 */
		DeviceInfo getDeviceInfo() const { return _info; }
	};

	/**
	 * Return full version info.
	 */ 
	Version getVersion();

	/**
	 * The API is not usable before calling this function.
	 */ 
	void init();

	/**
	 * Search for devices masked by the given device type specifier.
	 * @param type Filter search by class or type.
	 * @return List of detected devices. Instances are valid until the next call to enumerate.
	 */
	std::vector<EnumeratedDevice> enumerate(uint16_t vid, uint16_t pid);

	/**
	 * Free API resources, closes all devices and unloads all modules. To reuse the API again,
	 * init() must be called again.
	 */ 
	void cleanup();

	/**
	 * Set logging level (Off = 0, Error = 1, Warning = 2, Debug = 3), default is 2.
	 */
	void setLogLevel(int ll);

} // namespace beastlink
