/*
  CESYS Software License - Version 1.0 - January 1st, 2016

  Permission is hereby granted, free of charge, to any person or organization
  obtaining a copy of the software and accompanying documentation covered by
  this license (the "Software") to use, reproduce, display, distribute,
  execute, and transmit the Software, and to prepare derivative works of the
  Software, and to permit third-parties to whom the Software is furnished to
  do so, all subject to the following:

  The software must only be used to operate hardware manufactured by
  CESYS GmbH, Herzogenaurach.

  The copyright notices in the Software and this entire statement, including
  the above license grant, this restriction and the following disclaimer,
  must be included in all copies of the Software, in whole or in part, and
  all derivative works of the Software, unless such copies or derivative
  works are solely in the form of machine-executable object code generated by
  a source language processor.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
  SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
  FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.

  Written by Thomas Hoppe <thomas.hoppe@cesys.com>, 2018
*/

#ifndef	__BEASTLINK_API_H__
#define	__BEASTLINK_API_H__

#ifndef	WIN32
#	define BL_EXPORT
#else
#	ifdef	_BEAST_LIBRARY_
#		define BL_EXPORT	__declspec(dllexport)
#	else
#		define BL_EXPORT	__declspec(dllimport)
#	endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Return code for every function that may have an error condition.
 */
typedef int BL_RESULT;

/**
 * Data type to use for a device handle.
 */
typedef int BL_DEVICE_HANDLE;

#ifdef __cplusplus
	/**
	 * A buffer used for 0-terminated strings must be of this size, used in:
	 * BlGetLastErrorText(), BlSetUserId(), BlGetUserId(), BlGetDerivateInfo(),
	 * BlGetSerialNumber() and BlGetFirmwareVersion().
	 */
	const int BL_IF_BUFFER_SIZE = 16384;

	/**
	 * Returned by a function to indicate that no error has been occurred. All
	 * other return values are ID's that must be used to get the error reason
	 * from BlGetLastErrorText().
	 */
	const BL_RESULT BL_SUCCESS = 0;

	/**
	 * Handle value for invalid devices.
	 */
	const BL_DEVICE_HANDLE BL_INVALID_DEVICE = -1;

	/**
	 * This flag can be used in BlReadBlock() and BlWriteBlock() to specify transfer to a
	 * constant address. The opposite and *default* behavior is incremental address.
	 * Attention: For historical reasons, this flag is interpreted inverse inside the FPGA interface ->
	 * "address auto increment" is 0 if BL_FLAG_CONST_ADDR is specified.
	 */
	const int BL_FLAG_CONST_ADDR = 0x00000001;
#else
	/**
	 * A buffer used for 0-terminated strings must be of this size, used in:
	 * BlGetLastErrorText(), BlSetUserId(), BlGetUserId(), BlGetDerivateInfo(),
	 * BlGetSerialNumber() and BlGetFirmwareVersion().
	 */
#	define BL_IF_BUFFER_SIZE 16384

	/**
	 * Returned by a function to indicate that no error has been occurred. All
	 * other return values are ID's that must be used to get the error reason
	 * from BlGetLastErrorText().
	 */
#	define BL_SUCCESS 0

	/**
	 * Handle value for invalid devices.
	 */
#	define BL_INVALID_DEVICE -1

	/**
	 * This flag can be used in BlReadBlock() and BlWriteBlock() to specify transfer to a
	 * constant address. The opposite and *default* behavior is incremental address.
	 * Attention: For historical reasons, this flag is interpreted inverse inside the FPGA interface ->
	 * "address auto increment" is 0 if BL_FLAG_CONST_ADDR is specified.
	 */
#	define BL_FLAG_CONST_ADDR 0x00000001
#endif

#ifndef WIN32
#	ifdef	_BEAST_LIBRARY_
#		pragma GCC visibility push(default)
#	endif
#endif

/**
 * Return error text of last exception. The id value is the return code of the
 * failed function invocation. Every id is unique and can only be used once
 * calling this function (it is removed from the internal error collector
 * after the first access). buffer must be of size BL_IF_BUFFER_SIZE and
 * contains a 0-terminated string after the invocation.
 */
void BL_EXPORT BlGetLastErrorText(int id, char *buffer);

// *** Global static functions.

/**
 * Initialize API components.
 */
BL_RESULT BL_EXPORT BlInit();

/**
* Cleanup API components.
*/
BL_RESULT BL_EXPORT BlCleanup();

/**
 * Return full of BeastLink library.
 */
BL_RESULT BL_EXPORT BlGetVersion(int *major, int *minor, int *fix);

/**
 * Set logging level (Off = 0, Error = 1, Warning = 2, Debug = 3), default is 2.
 */
BL_RESULT BL_EXPORT BlSetLogLevel(int ll);

/**
 * Enumerates devices with the given VID and PID. count returns number of devices found.
 */
BL_RESULT BL_EXPORT BlEnumerate(unsigned short vid, unsigned short pid, int *count);

// *** Generic access.

/**
 * Access device. enumId must be in range 0 to count (count returned by BlEnumerate()). A new call
 * to BlEnumerate invalidates the numbers for this id. The returned handle must be used for
 * every subsequent communication with the device. An opened device isn't reported in
 * any subsequent enumeration (system wide), until it's closed.
 */
BL_RESULT BL_EXPORT BlOpen(int enumId, BL_DEVICE_HANDLE *handle);

/**
 * Free access to device. After this call the device is available in new enumerations.
 */
BL_RESULT BL_EXPORT BlClose(BL_DEVICE_HANDLE handle);

// *** Data transfer functions.

/**
 * Read 32 bit register value. handle is returned from BlOpen(). reg specifies the register
 * in FPGA address space. result contains the value read from device.
 */
BL_RESULT BL_EXPORT BlReadRegister(BL_DEVICE_HANDLE handle, unsigned int reg, unsigned int *result);

/**
 * Write 32 bit register value. handle is returned from BlOpen(). reg specifies the register
 * in FPGA address space. value will be the new value.
 */
BL_RESULT BL_EXPORT BlWriteRegister(BL_DEVICE_HANDLE handle, unsigned int reg, unsigned int value);

/**
 * Read data block. handle is returned from BlOpen(). addr is the address in FPGA address space to which
 * the data is transferred. buffer is the pointer to the data the should be send. size is the count
 * of bytes to transfer. flags can be (see their individual documentation):
 *   - BL_FLAG_CONST_ADDR
 */
BL_RESULT BL_EXPORT BlReadBlock(BL_DEVICE_HANDLE handle, unsigned int addr, void *buffer, int size, int flags);

/**
 * Write data block. handle is returned from BlOpen(). addr is the address in FPGA address space from which
 * the data is transferred. buffer is the pointer to which received data is stored. size is the count
 * of bytes to transfer. flags can be (see their individual documentation):
 *   - BL_FLAG_CONST_ADDR
 */
BL_RESULT BL_EXPORT BlWriteBlock(BL_DEVICE_HANDLE handle, unsigned int addr, void *buffer, int size, int flags);

// *** FPGA specific.

/**
 * Pulse FPGA reset line (the I/O that is chosen to be reset). Useful to bring the FPGA
 * in a defined state. Implicitly done after FPGA programming. handle is returned from BlOpen().
 */
BL_RESULT BL_EXPORT BlResetFpga(BL_DEVICE_HANDLE handle);

/**
 * Configure FPGA with given design in file. handle is returned from BlOpen().
 * Loads a .bin from as specified by fileName and programs the FPGA using this design.
 * Subsequently pulses the reset line.
 */
BL_RESULT BL_EXPORT BlProgramFpgaFromBin(BL_DEVICE_HANDLE handle, const char *fileName);

/**
 * Configure FPGA with given design in memory. Useful when compiling FPGA designs into the executable.
 * BoardManager creates sources that can be used this way.
 * The reset line is pulsed subsequently.
 */
BL_RESULT BL_EXPORT BlProgramFpgaFromMemory(BL_DEVICE_HANDLE handle, void *buffer, int size, int compressed);

// *** Configuration.

/**
 * Configure timeout until a pending transfer is stopped and defined as failed.
 * This timeout is used for all transfers to and from device.
 * handle is returned from BlOpen(). timeout must be specified in milliseconds.
 */
BL_RESULT BL_EXPORT BlSetTimeout(BL_DEVICE_HANDLE handle, unsigned int timeout);

// *** User ID.

/**
 * Set user ID. Writes the 0 - terminated string in buffer to the device EEPROM.
 * This ID is intended to differentiate multiple devices of the same type connected
 * to one host. The ID can contain up to 63 ASCII characters.
 */
BL_RESULT BL_EXPORT BlSetUserId(BL_DEVICE_HANDLE handle, char *buffer);

/**
 * Return user ID from device EEPROM as 0 - terminated string. See explanation in
 * BlSetUserId().
 */
BL_RESULT BL_EXPORT BlGetUserId(BL_DEVICE_HANDLE handle, char *buffer);

// *** Derivate info.

/**
 * Return derivate info string. handle is returned from BlOpen(). buffer must be of
 * size BL_IF_BUFFER_SIZE.
 * Some devices are available in different versions (different FPGA's, different peripherals).
 * The returned 0 - terminated string contains a textual information about the specific
 * derivate. For programmatically differentiation, BlGetDerivateId() is better suited.
 */
BL_RESULT BL_EXPORT BlGetDerivateInfo(BL_DEVICE_HANDLE handle, char *buffer);

/**
 * Return derivate id. handle is returned from BlOpen(). This ID is unique for every
 * derivate of a specific device type. Existing derivates can be found in the documentation
 * of the specific device type.
 */
BL_RESULT BL_EXPORT BlGetDerivateId(BL_DEVICE_HANDLE handle, unsigned int *derivateId);

// *** General device info.

/**
 * Return serial number as 0 - terminated string. Not supported by all device types / instances.
 * handle is returned from BlOpen(). buffer must be of size BL_IF_BUFFER_SIZE.
 */
BL_RESULT BL_EXPORT BlGetSerialNumber(BL_DEVICE_HANDLE handle, char *buffer);

/**
 * Return firmware version. handle is returned from BlOpen().
 * buffer must be of size BL_IF_BUFFER_SIZE.
 */
BL_RESULT BL_EXPORT BlGetFirmwareVersion(BL_DEVICE_HANDLE handle, int *major, int *minor, int *fix);

#ifndef WIN32
#	ifdef	_BEAST_LIBRARY_
#		pragma GCC visibility pop
#	endif
#endif

#ifdef __cplusplus
} // extern "C"
#endif

#endif // __BEASTLINK_API_H__
